!< 
!-------------------------------------------------------------
! . . Exact_r4.x
! . . Program that calculates exact rays based on the theory 
! . . of Terry L. Foreman in 
! . . An exact ray theoretical formulation of the Helmholtz
! . . equation, J. Acoust. Soc. Am. 86(1), July 1989
!
! USAGE:
! . . < in.H Exact_r1.x par=parfile.p > out.H
! . .
! . . where parfile.p contains
! . . 
! . . xsp = x location of shotpoint
! . . R = Radius of circle to start about xsp
! . . begang = starting angle (in degrees) at Radius R about shot point xsp 
! . . endang = finishing angle (in degrees) at Radius R about shot point xsp 
! . . maxnum = maximum number of steps with numerical integrator
! . . eps = (0) a small number to prevent a division by zero
! . . 
!& end of self-documentation
!-------------------------------------------------------------
program Exact_r4
  use SEP
  use exact_r4_mod 
  implicit none
  integer                                :: nz,nx,nw,maxnum,i2,nray
  integer                                :: iray,it,wantP,ic,cf
  integer,  dimension(:)                 :: ixx,izz,ntodo,angtodo,ib1,ib2,ib3
  real(KIND=8)                           :: pi
  real                                   :: dtau,begang,endang,dz,dx,dw,oz,ox
  real                                   :: ow,eps,xsp,R,mxint,mnint
  real(KIND=8),dimension(:),allocatable  :: xtodo,ztodo,x,z,xang,zang
  real        ,dimension(:,:),allocatable:: vel_in
  real(kind=8),dimension(:,:),allocatable:: vel
  complex,  dimension(:,:),allocatable   :: Rayout,data
  logical                                :: adj,add,fromSP
  call sep_init()
  call from_history("n1",nz             ) 
  call from_history("n2",nx             )  
  call from_history("n3",nw             )  
  call from_history("d1",dz             ) 
  call from_history("d2",dx             )  
  call from_history("d3",dw             )  
  call from_history("o1",oz             ) 
  call from_history("o2",ox             )  
  call from_history("o3",ow             )
  call from_param("adj", adj, .false.   )
  call from_param("add", add, .false.   ) 
  call from_param("eps", eps, 0.000001  )
  call from_param("maxnum",maxnum,700   )
  call from_param("xsp", xsp            )
  call from_param("R",   R              )
  call from_param("begang",begang,-45.  ) 
  call from_param("endang",endang,45.   )
  call from_param("nray",nray,100       )
  call from_param("dtau",dtau           )
  call from_param("fromSP",fromSP,.true.)
  call from_param("mxint",mxint,5.      )
  call from_param("mnint",mnint,0.05    )
  call from_param("wantP",wantP,0       )
  allocate( xtodo(nray),ztodo(nray)     )
  allocate( x(nx),z(nz),ntodo(nray)     )
  allocate( Rayout( maxnum,nx)          )
  allocate( ixx(nx),izz(nz),data(nz,nx) )
  allocate( vel(nz,nx),vel_in(nz,nx)    )
  call sep_read( data                   )
  call sep_read( vel_in,"vel"           )
  vel=dble(vel_in)
  call exact_mod_init(nz,nx,nw,dz,dx,dw,oz,ox,ow,dtau,mxint,fromSP,xsp,R,eps,mnint)
  ixx(1:nx) = (/ (i2,i2=1,nx) /);  izz(1:nz) = (/ (i2,i2=1,nz) /) 
  x=(ixx-1)*dx+ox               ;  z=(izz-1)*dz+oz
  ntodo = (/ (iray, iray=0,nray-1) /)  
  if (wantP .eq. 1 ) then
     allocate( xang(nray),zang(nray) )
     xang= x
     zang= dz
     write(0,*) ' USING PLANE WAVES '
  else if (wantP .eq. 2) then
     cf=30
     allocate( xang( (nx+2*nz)/cf),zang( (nx+2*nz)/cf) )         
     do ic=1,nz/cf 
        xang(ic)=x(nx/cf)
        zang(ic)=z(cf*ic)
     end do
     do ic=1,nx/cf
        xang(ic+nz/cf)=x(cf*ic)
        zang(ic+nz/cf)=z(nz-cf)
     end do
     do ic=1,nz/cf-1
        xang(ic+ (nx+nz)/cf)=x((cf-1)*nx/cf)
        zang(ic+ (nx+nz)/cf)=z(nz-cf*ic)
     end do
  else  
     allocate( xang(nray),zang(nray) )
     pi=2.*asin(1.)
     xang=R*cos( (endang-begang)*pi/(180*(nray-1))*ntodo+begang*pi/180.0+pi/2.0) + xsp
     zang=R*sin( (endang-begang)*pi/(180*(nray-1))*ntodo+begang*pi/180.0+pi/2.0) 
  end if
  call Run_exact_interp(vel,data,Rayout,adj,add,maxnum,x,z,xang,zang,it)
  call write_on_baby(Rayout,maxnum,it,endang,begang)
  deallocate(Rayout,xang,zang,data,vel,x,z)
end program Exact_r4
