!< 
!-------------------------------------------------------------
! . . Exact.x
! . . Program that calculates exact rays based on the theory 
! . . of Terry L. Foreman in 
! . . An exact ray theoretical formulation of the Helmholtz
! . . equation, J. Acoust. Soc. Am. 86(1), July 1989
!
! USAGE:
! . . < in.H Exact.x par=parfile.p > out.H
! . .
! . . where parfile.p contains
! . . 
! . . xsp = x location of shotpoint
! . . R = Radius of circle to start about xsp
! . . begang = starting angle (in degrees) at Radius R about shot point xsp 
! . . endang = finishing angle (in degrees) at Radius R about shot point xsp 
! . . maxnum = maximum number of steps with numerical integrator
! . . eps = (0) a small number to prevent a division by zero
! . . 
!& end of self-documentation
!-------------------------------------------------------------
program Exact
  use SEP
  use exact_mod 
  implicit none
  integer                                :: nz,nx,nw,maxnum,i2,nray,iray
  integer, dimension(:)                  :: ixx,izz,ntodo
  real                                   :: dz,dx,dw,oz,ox,ow,eps,xsp,R,pi,begang,endang,ds
  real, dimension(:),allocatable         :: xtodo,ztodo,x,z
  complex, allocatable, dimension(:,:,:) :: Rayout
  logical                                :: adj,add,fromSP,wantP
  call sep_init()
  call from_history("n1",nz           ) 
  call from_history("n2",nx           )  
  call from_history("n3",nw           )  
  call from_history("d1",dz           ) 
  call from_history("d2",dx           )  
  call from_history("d3",dw           )  
  call from_history("o1",oz           ) 
  call from_history("o2",ox           )  
  call from_history("o3",ow           )
  call from_param("adj", adj, .false. )
  call from_param("add", add, .false. ) 
  call from_param("eps", eps, 0.000001)
  call from_param("maxnum",maxnum,700 )
  call from_param("xsp", xsp          )
  call from_param("R",   R            )
  call from_param("begang",begang,0.  ) 
  call from_param("endang",endang,180.)
  call from_param("nray",nray,100     )
  call from_param("ds",ds,max(dx,dz)  )
  call from_param("fromSP",fromSP,.true.)
  call from_param("wantP", wantP,.false.)
  allocate(xtodo(nray),ztodo(nray))
  allocate(x(nx),z(nz))
  allocate( Rayout( maxnum,nray,nw) )
  allocate( ixx(nx), izz(nz) )
  allocate( ntodo(nray) )
  !! . . Initialize module
  if (fromSP .eq. .false.) then
     ds=-ds 
     write(0,*) 'Flipped distance'
  endif
  call exact_mod_init(nz,nx,nw,dz,dx,dw,oz,ox,ow,ds)
  pi=3.1415926

  write(0,*) 'USING beginning and end angles, and num rays of: ',begang,endang,nray
  write(0,*) 'step size is: ',ds

  !! . . Input model dimensions
  ixx(1:nx) = (/ (i2,i2=1, nx) /) 
  izz(1:nz) = (/ (i2,i2=1, nz) /) 
  x=(ixx-1)*dx+ox 
  z=(izz-1)*dz+oz 

  !! . . User defined window
  ntodo = (/ (iray, iray=0,nray-1) /)  
  if (wantP .eq. true) then
     xtodo=x
     ztodo=3*dz 
  else
     xtodo=R*cos( (endang-begang)*pi/(180*nray)*ntodo+begang*pi/180+pi/2) + xsp
     ztodo=R*sin( (endang-begang)*pi/(180*nray)*ntodo+begang*pi/180+pi/2) 
  endif

  !! . . Run exact program 
  call Run_exact(Rayout,adj,add,eps,maxnum,x,z,xtodo,ztodo)

  !! . . Write out
  call to_history("n1",maxnum               ) 
  call to_history("n2",nray                 )  
  call to_history("n3",nw                   )  
  call to_history("d1",ds/1000              ) 
  call to_history("d2",1                    )  
  call to_history("d3",dw                   )  
  call to_history("o1",0                    ) 
  call to_history("o2",begang               )  
  call to_history("o3",dw                   )  
  call hclose()
  call sep_write(Rayout)
end program Exact
