!< 
!-------------------------------------------------------------
! . . Stepping.f90
! . . Program that calculates exact rays based on the theory 
! . . of Terry L. Foreman in 
! . . An exact ray theoretical formulation of the Helmholtz
! . . equation, J. Acoust. Soc. Am. 86(1), July 1989
!
! USAGE:
! . . < in.H Exact_r1.x par=parfile.p > out.H
! . .
! . . where parfile.p contains
! . . 
! . . maxnum = maximum number of steps with numerical integrator
! . . eps = (0) a small number to prevent a division by zero
! . . 
!& end of self-documentation
!-------------------------------------------------------------
program Stepping
  use SEP
  use stepping_mod 
  implicit none
  integer                       :: nvz,nx,nw,nvx,maxnum,itau
  integer                       :: ang_win,ang_num,abcd_win,abcd_num
  real(KIND=8)                  :: dvx, dvz, dx, dw, ovx, ovz, ox, ow
  real                          :: dvx1,dvz1,dx1,dw1,ovx1,ovz1,ox1,ow1
  real(KIND=8)                  :: eps, gridang, dtau
  real                          :: eps1,gridang1,dtau1
  real(KIND=8)   ,dimension(:  ),allocatable :: x,z,a,b,c,d,s
  real(KIND=8)   ,dimension(:,:),allocatable :: xsta,zsta,ssta
  real           ,dimension(:,:),allocatable :: vel_in
  real(kind=8)   ,dimension(:,:),allocatable :: vel
  complex        ,dimension(:,:),allocatable :: rays,grid_in
  complex        ,dimension(:)  ,allocatable :: data_in,raywrite,imgwrite
  complex(KIND=8),dimension(:)  ,allocatable :: data,img,imgout
  complex(KIND=8),dimension(:,:),allocatable :: imgsta,grid
  call sep_init()
  call from_history("n2",nw             )  
  call from_history("d2",dw1             );dw=dble(dw1)  
  call from_history("o2",ow1             );ow=dble(ow1)
  call from_history("n1",nx             )  
  call from_history("d1",dx1             );dx=dble(dx1) 
  call from_history("o1",ox1             );ox=dble(ox1) 
  call from_aux("vel","n1",nvz          ) 
  call from_aux("vel","d1",dvz1          );dvz=dble(dvz1)
  call from_aux("vel","o1",ovz1          );ovz=dble(ovz1) 
  call from_aux("vel","n2",nvx          ) 
  call from_aux("vel","d2",dvx1          ) ;dvx=dble(dvx1)
  call from_aux("vel","o2",ovx1          ) ;ovx=dble(ovx1) 
  call from_param("eps", eps1, 0.000001  );eps=dble(eps1) 
  call from_param("maxnum",maxnum,700   ) 
  call from_param("dtau",dtau1           ); dtau=dble(dtau1) 
  call from_param("gridang",gridang1,0.  ); gridang=dble(gridang1) 
  call from_param("ang_num",ang_num,5    )
  call from_param("ang_win",ang_win,5    )
  call from_param("abcd_num",abcd_num,5    )
  call from_param("abcd_win",abcd_win,5    )
  allocate( xsta(4,nx),zsta(4,nx),ssta(4,nx) )
  allocate( x(nvx),z(nvz),imgsta(4,nx)    )
  allocate( img(nx),imgout(nx)            )
  allocate( data(nx),data_in(nx)          )
  allocate( grid(4,nx),grid_in(4,nx)      )
  allocate( vel(nvz,nvx),vel_in(nvz,nvx)  )
  allocate( a(nx),b(nx),c(nx),d(nx),s(nx) )
  allocate( rays(maxnum,nx)               )
  allocate( raywrite(nx),imgwrite(nx)     )
  call sep_read( data_in                  )
  call sep_read( grid_in,"grid"           )
  call sep_read( vel_in ,"vel "           )
  data= dble(data_in)
  grid= dble(grid_in)
  vel = dble(vel_in )
  rays(1:4,:) = grid_in
  xsta(1:4,:) = dble( real(grid_in))
  zsta(1:4,:) = dble(aimag(grid_in))
  deallocate(vel_in,data_in )

  !! . . Set up variables
  call stepping_init1(nvx,nvz,nx,nw,dvx,dvz,dx,dw,ovx,ovz,ox,ow,&
    dtau,eps,gridang,ang_win,ang_num,abcd_win,abcd_num)
  call write_out(maxnum)

  !! . . Get model and starting grid locations
  call model_points(x,z)
  call get_start_slow(x,z,xsta,zsta,vel,ssta)

  do itau=1,4
     raywrite = rays(itau,:)
     call sep_write( raywrite,"rays.H",nx,8)
  end do

  !! STEP 1 . . First W.F. quantity
  imgsta(1,:)=data
  imgwrite=data
  call sep_write( imgwrite )
  
  !! STEP 2 . . Cartesian
  img=data
  itau=1
  call Ray_spreading(xsta,zsta,a,b,c,d,s)
  do itau=2,4
  call Ray_spreading(xsta,zsta,a,b,c,d,s)
     s=ssta(itau,:)
     call Jmigrc15(img,img,a,b,c,d,s)
     imgsta(itau,:)=img
     imgwrite=img
     call sep_write( imgwrite )
  end do

  !! STEP 3 to maxnum
  do itau=5,maxnum
write(0,*) 'Working on step ',itau, ' of ',maxnum

     !! . . Calculate new step
     call Step_calculation(xsta,zsta,vel,imgsta,x,z,s)
     imgsta(1:3,:)=imgsta(2:4,:)

     !! . . Calculate rays
     call Ray_spreading(xsta,zsta,a,b,c,d,s)
     raywrite=cmplx(xsta(4,:),zsta(4,:))
     call sep_write( raywrite,"rays.H",nx,8)

     !! . . Calculate FD's
     call Jmigrc15(img,img,a,b,c,d,s)
     imgsta(4,:)=img
     imgwrite=img
     call sep_write( imgwrite )
  end do

  deallocate(xsta,zsta,data,vel,x,z,a,b,c,d,s,raywrite,imgwrite,imgout)
end program Stepping
