!< 
!-------------------------------------------------------------
! . . Toporays.x
! . . Program that calculates exact rays based on the theory 
! . . of Terry L. Foreman in 
! . . An exact ray theoretical formulation of the Helmholtz
! . . equation, J. Acoust. Soc. Am. 86(1), July 1989
!
! USAGE:
! . . < in.H Exact_r1.x par=parfile.p > out.H
! . .
! . . where parfile.p contains
! . . 
! . . xsp = x location of shotpoint
! . . R = Radius of circle to start about xsp
! . . begang = starting angle (in degrees) at Radius R about shot point xsp 
! . . endang = finishing angle (in degrees) at Radius R about shot point xsp 
! . . maxnum = maximum number of steps with numerical integrator
! . . eps = (0) a small number to prevent a division by zero
! . . 
!& end of self-documentation
!-------------------------------------------------------------
program Toporays
  use SEP
  use toporays_mod 
  implicit none
  integer                                    :: nz,nx,nw,maxnum,nray,nvx,nvz,nwx,nwz
  real                                       :: dx1,dw1,ox1,ow1,dvz1,dvx1,ovz1,ovx1,dz1,oz1
  real(KIND=8)                               :: dx ,dw ,ox ,ow ,dvz ,dvx ,ovz ,ovx, dz, oz
  real                                       :: dwz1,dwx1,owx1,owz1
  real(KIND=8)                               :: dwz ,dwx ,owx ,owz
  real                                       :: eps1,dtau1
  real(KIND=8)                               :: eps, dtau 
  complex        ,dimension(:  ),allocatable :: start_in
  real(KIND=8)   ,dimension(:  ),allocatable :: xsta,zsta,x,z
  real           ,dimension(:,:),allocatable :: vel_in,data_in
  real(KIND=8)   ,dimension(:,:),allocatable :: vel,data
  complex        ,dimension(:,:),allocatable :: Rayout
  complex(KIND=8),dimension(:,:),allocatable :: rays

  call sep_init()
  call from_history("n1",nx             )  
  call from_history("n2",nw             )
  call from_history("d1",dx1            );dx=dble(dx1)
  call from_history("d2",dw1            );dw=dble(dw1) 
  call from_history("o1",ox1            );ox=dble(ox1)  
  call from_history("o2",ow1            );ow=dble(ow1) 

  call from_aux("vel","n1",nvz          ) 
  call from_aux("vel","d1",dvz1         );dvz=dble(dvz1)
  call from_aux("vel","o1",ovz1         );ovz=dble(ovz1) 
  call from_aux("vel","n2",nvx          ) 
  call from_aux("vel","d2",dvx1         ) ;dvx=dble(dvx1)
  call from_aux("vel","o2",ovx1         ) ;ovx=dble(ovx1) 

  call from_aux("wf","n1",nwz           ) 
  call from_aux("wf","d1",dwz1          );dwz=dble(dwz1)
  call from_aux("wf","o1",owz1          );owz=dble(owz1) 
  call from_aux("wf","n2",nwx           ) 
  call from_aux("wf","d2",dwx1          ) ;dwx=dble(dwx1)
  call from_aux("wf","o2",owx1          ) ;owx=dble(owx1) 

  call from_param("eps", eps1, 0.000001 );eps=dble(eps1)
  call from_param("maxnum",maxnum,700   )
  call from_param("nray",nray,100       )
  call from_param("dtau",dtau1          );dtau=dble(dtau1)

  call to_history("n1",maxnum)
  call to_history("n2",nx    )
  call to_history("d1",dtau1  )
  call to_history("d2",dx1   )
  call to_history("o1",0.    )
  call to_history("o2",ox1    )
  call to_history("esize",8  )
  call hclose()

  allocate( vel_in(nvz,nvx),  vel(nvz,nvx)   )
  allocate( data_in(nwz,nwx), data(nwz,nwx)  )
  allocate( start_in(nx),xsta(nx),zsta(nx)   )
  allocate( x(nx),z(nz)                      )
  allocate( Rayout(maxnum,nx),rays(maxnum,nx))

  call sep_read( vel_in,"vel"                )
  call sep_read( data_in,"wf"                )
  call sep_read( start_in                    )
  vel   = dble( vel_in)
  data  = dble(data_in)
  xsta  = dble( real (start_in) )  
  zsta  = dble( aimag(start_in) )

  call toporays_init(nz,nx,nw,dz,dx,dw,oz,ox,ow,&
      nvx,nvz,dvx,dvz,ovx,ovz,nwx,nwz,dwx,dwz,owx,owz,dtau,eps,maxnum)
  call model_points(x,z)
  call toporays(vel,data,rays,x,z,xsta,zsta)
  Rayout=rays
  call sep_write(Rayout)

  deallocate(Rayout,xsta,zsta,data,vel,x,z)
end program Toporays
